package com.bergerkiller.bukkit.nolagg;

import java.util.List;

import net.minecraft.server.Packet;
import net.minecraft.server.Packet51MapChunk;
import net.minecraft.server.Packet53BlockChange;

import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Item;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPhysicsEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.block.LeavesDecayEvent;
import org.bukkit.event.entity.CreatureSpawnEvent;
import org.bukkit.event.entity.EntityDeathEvent;
import org.bukkit.event.entity.ItemSpawnEvent;
import org.bukkit.event.entity.CreatureSpawnEvent.SpawnReason;
import org.bukkit.event.player.PlayerPickupItemEvent;
import org.bukkit.event.world.ChunkLoadEvent;
import org.bukkit.event.world.ChunkPopulateEvent;
import org.bukkit.event.world.ChunkUnloadEvent;
import org.bukkit.event.world.WorldLoadEvent;
import org.bukkit.event.world.WorldUnloadEvent;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.common.Task;
import com.bergerkiller.bukkit.common.utils.PacketUtil;
import com.bergerkiller.bukkit.common.utils.WorldUtil;
import com.bergerkiller.bukkit.nolagg.meta.ChunkMetaData;
import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;
import com.bergerkiller.bukkit.nolagg.spawnlimit.EntityLimiter;
import com.bergerkiller.bukkit.nolagg.spawnlimit.SpawnHandler;

public class NLListener implements Listener {
	
	public static long loadedChunks = 0;
	public static long unloadedChunks = 0;
	public static long genChunks = 0;
	public static boolean wasChunkLightingFixed = false;
			
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onLeavesDecay(LeavesDecayEvent event) {
		if (event.isCancelled()) return;
		fixColumn(event.getBlock(), 2);
	}
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onBlockBreak(BlockBreakEvent event) {
		if (event.isCancelled()) return;
		fixColumn(event.getBlock());
	}
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onBlockPlace(BlockPlaceEvent event) {
		if (event.isCancelled()) return;
		fixColumn(event.getBlock(), 1);
	}
	@EventHandler(priority = EventPriority.HIGHEST)
	public void onBlockPhysics(BlockPhysicsEvent event) {
		if (event.isCancelled()) return;
		Material type = event.getBlock().getType();
		if (NoLagg.useSpawnLimits) {
			if (type == Material.GRAVEL || type == Material.SAND) {
				EntityLimiter limiter = SpawnHandler.GENERALHANDLER.getEntityLimits(event.getBlock().getWorld(), "falling" + type.toString());
				if (limiter != null) {
					if (limiter.canSpawn()) {
						limiter.spawn();
					} else {
						event.setCancelled(true);
					}
				}
			}
		}
	}
	
	@EventHandler(priority = EventPriority.LOWEST)
	public void onItemSpawn(ItemSpawnEvent event) {
		if (!ThreadCheck.check("")) return;
		if (event.isCancelled()) return;
		if (NoLagg.useSpawnLimits) {
			PerformanceMonitor.entitySpawnLimit.start();
			if (!SpawnHandler.getEntityLimits(event.getEntity()).canSpawn()) {
				event.setCancelled(true);
			}
			PerformanceMonitor.entitySpawnLimit.next();
		}
		if (!event.isCancelled() && NoLagg.bufferItems) {
			if (!WorldMetaData.handleItemSpawn((Item) event.getEntity(), false)) {
				event.setCancelled(true);
			}
		}
	}
	
	@EventHandler(priority = EventPriority.LOWEST)
	public void onCreatureSpawn(CreatureSpawnEvent event) {
		if (!ThreadCheck.check("")) return;
		if (NoLagg.useSpawnLimits) {
			PerformanceMonitor.entitySpawnLimit.start();
			if (event.getSpawnReason() == SpawnReason.CUSTOM) {
				SpawnHandler.ignoreSpawn(event.getEntity());
			} else {
				if (event.getSpawnReason() == SpawnReason.SPAWNER) {
					SpawnHandler.mobSpawnerSpawned(event.getEntity());
				}
				if (!SpawnHandler.getEntityLimits(event.getEntity()).canSpawn()) {
					event.setCancelled(true);
				}
			}
			PerformanceMonitor.entitySpawnLimit.next();
		}
	}
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void onCreatureSpawnMonitor(CreatureSpawnEvent event) {
		if (!ThreadCheck.check("")) return;
		if (NoLagg.useSpawnLimits) {
			PerformanceMonitor.entitySpawnLimit.start();
			if (!event.isCancelled()) {
				SpawnHandler.getEntityLimits(event.getEntity()).spawn();
			}
			PerformanceMonitor.entitySpawnLimit.next();
		}
	}
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void onItemSpawnMonitor(ItemSpawnEvent event) {
		if (!ThreadCheck.check("")) return;
		if (NoLagg.useSpawnLimits) {
			PerformanceMonitor.entitySpawnLimit.start();
			if (!event.isCancelled()) {
				SpawnHandler.getEntityLimits(event.getEntity()).spawn();
			}
			PerformanceMonitor.entitySpawnLimit.next();
		}
	}
	
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void onEntityDeath(EntityDeathEvent event) {
		if (!ThreadCheck.check("")) return;
		PerformanceMonitor.entitySpawnLimit.start();
		SpawnHandler.handleDespawn(event.getEntity());
		PerformanceMonitor.entitySpawnLimit.next();
	}
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void onPlayerPickupItem(PlayerPickupItemEvent event) {
		if (!event.isCancelled()) {
			WorldMetaData.removeSpawnedItem(event.getItem());
		}
	}
			
	@EventHandler(priority = EventPriority.MONITOR)
	public void onChunkLoad(ChunkLoadEvent event) {
		if (!ThreadCheck.check("")) return;
		PerformanceMonitor.chunkUpdate.start();
		ChunkMetaData data = ChunkMetaData.get(event.getChunk());
		data.touch();
		data.update();
		PerformanceMonitor.chunkUpdate.next();
		loadedChunks++;
		if (NoLagg.useSpawnLimits) {
			PerformanceMonitor.entitySpawnLimit.start();
			//properly clear entities in this chunk if needed
			for (Entity e : event.getChunk().getEntities()) {
				if (!SpawnHandler.getEntityLimits(e).handleSpawn()) {
					e.remove();
				}
			}
			PerformanceMonitor.entitySpawnLimit.next();
		}
	}
	
	@EventHandler(priority = EventPriority.LOWEST)
	public void onChunkUnload(ChunkUnloadEvent event) {
		if (event.isCancelled()) return;
		if (!ThreadCheck.check("")) return;
		PerformanceMonitor.chunkUpdate.start();
		ChunkMetaData data = ChunkMetaData.get(event.getChunk());
		if (!data.isUnloadable(false) || !data.isExpired()) {
			event.setCancelled(true);
			data.update();
		}
		PerformanceMonitor.chunkUpdate.next();
	}
		
	@EventHandler(priority = EventPriority.MONITOR)
	public void onWorldLoad(WorldLoadEvent event) {
		if (!ThreadCheck.check("")) return;
		WorldMetaData.get(event.getWorld());
		AutoSaveChanger.change(event.getWorld());
	}
	
	@EventHandler(priority = EventPriority.MONITOR)
	public void onWorldUnload(WorldUnloadEvent event) {
		if (!ThreadCheck.check("")) return;
		WorldMetaData.remove(event.getWorld());
	}
	
	@EventHandler(priority = EventPriority.LOWEST)
	public void onChunkPopulate(ChunkPopulateEvent event) {
		genChunks++;
		if (!NoLagg.autoFixLighting) return;
		OperationScheduler.start(new ChunkFixLightingOperation(event.getChunk()));
		wasChunkLightingFixed = true;
	}
		
	public boolean isTransparent(Material type) {
		return isTransparent(type.getId());
	}
	public boolean isTransparent(int typeid) {
		return !net.minecraft.server.Block.o[typeid];
	}
	public boolean isSemiTransparent(Material mat) {
		switch (mat) {
		case ICE : return true;
		case LEAVES : return true;
		default: return false;
		}
	}
	
	public void fixColumn(Block start) {
		fixColumn(start, 1);
	}
	public void fixColumn(Block start, int delay) {
		fixColumn(start, start.getType(), delay);
	}
	public void fixColumn(final Block start, final Material type, int delay) {
		if (!NoLagg.autoFixLighting) return;
		new Task(NoLagg.plugin){
			public void run() {
				fixColumn(start, type);
			}
		}.start(delay);
	}
	public void fixColumn(final Block start, final Material type) {
		final List<Player> players = start.getWorld().getPlayers();
		if (players.isEmpty()) return;
		fixColumn(players, start, type);
	}
	public void fixColumn(final List<Player> players, Block start, Material type) {
		while (true) {
			while (isSemiTransparent(type)) {
				start = start.getRelative(BlockFace.DOWN);
				if (start.getY() < 2) return;
				type = start.getType();
			}
			//air?
			int height = 0;
			while (isTransparent(type) || isSemiTransparent(type)) {
				start = start.getRelative(BlockFace.DOWN);
				if (start.getY() < 2) break;
				type = start.getType();
				height++;
			}
			if (height == 0) return;
			//send packet for this stack of blocks
			Packet packet;
			if (height == 1) {
				packet = new Packet53BlockChange(start.getX(), start.getY() + 1, start.getZ(), WorldUtil.getNative(start.getWorld()));
			} else {
				packet = new Packet51MapChunk(start.getX(), start.getY() + 1, start.getZ(), 1, height, 1, WorldUtil.getNative(start.getWorld()));
			}
			PacketUtil.broadcastBlockPacket(start, packet, true);
		}
	}
	
}
